-module(erl_parse).
-define(THIS_MODULE, erl_parse).
-export([parse/1, parse_and_scan/1, format_error/1]).

%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

-export([parse_form/1,parse_exprs/1,parse_term/1]).
-export([normalise/1,abstract/1,tokens/1,tokens/2]).
-export([abstract/2, package_segments/1]).
-export([inop_prec/1,preop_prec/1,func_prec/0,max_prec/0]).

-compile(inline).

%% mkop(Op, Arg) -> {op,Line,Op,Arg}.
%% mkop(Left, Op, Right) -> {op,Line,Op,Left,Right}.

mkop(L, {Op,Pos}, R) -> {op,Pos,Op,L,R}.

mkop({Op,Pos}, A) -> {op,Pos,Op,A}.

%% keep track of line info in tokens
line(Tup) -> element(2, Tup).

%% Entry points compatible to old erl_parse.
%% These really suck and are only here until Calle gets multiple
%% entry points working.

parse_form(Tokens) ->
    parse(Tokens).

parse_exprs(Tokens) ->
    case parse([{atom,0,f},{'(',0},{')',0},{'->',0}|Tokens]) of
	{ok,{function,Lf,f,0,[{clause,Lc,[],[],Exprs}]}} ->
	    {ok,Exprs};
	{error,E} -> {error,E}
    end.

parse_term(Tokens) ->
    case parse([{atom,0,f},{'(',0},{')',0},{'->',0}|Tokens]) of
	{ok,{function,Lf,f,0,[{clause,Lc,[],[],[Expr]}]}} ->
	    case catch normalise(Expr) of
		{'EXIT',R} ->
		    {error,{line(Expr),?MODULE,"bad term"}};
		Term -> {ok,Term}
	    end;
	{ok,{function,Lf,f,0,[{clause,Lc,[],[],[E1,E2|Es]}]}} ->
	    {error,{line(E2),?MODULE,"bad term"}};
	{error,E} -> {error,E}
    end.

%% build_attribute(AttrName, AttrValue) ->
%%	{attribute,Line,module,Module}
%%	{attribute,Line,export,Exports}
%%	{attribute,Line,import,Imports}
%%	{attribute,Line,record,{Name,Inits}}
%%	{attribute,Line,file,{Name,Line}}
%%	{attribute,Line,Name,Val}

build_attribute({atom,La,module}, Val) ->
    case Val of
	[{atom,Lm,Module}] ->
	    {attribute,La,module,Module};
	[Name] ->
	    case package_segments(Name) of
		error ->
		    error_bad_decl(La, module);
		Module ->
		    {attribute,La,module,Module}
	    end;
	Other ->
	    error_bad_decl(La, module)
    end;
build_attribute({atom,La,export}, Val) ->
    case Val of
	[ExpList] ->
	    {attribute,La,export,farity_list(ExpList)};
	Other -> error_bad_decl(La, export)
    end;
build_attribute({atom,La,import}, Val) ->
    case Val of
	[Name] ->
	    case package_segments(Name) of
		error ->
		    error_bad_decl(La, import);
		Module ->
		    {attribute,La,import,Module}
	    end;
	[{atom,Lm,Mod},ImpList] ->
	    {attribute,La,import,{Mod,farity_list(ImpList)}};
	[Name, ImpList] ->
	    case package_segments(Name) of
		error ->
		    error_bad_decl(La, import);
		Module ->
		    {attribute,La,import,{Module,farity_list(ImpList)}}
	    end;
	Other -> error_bad_decl(La, import)
    end;
build_attribute({atom,La,record}, Val) ->
    case Val of
	[{atom,Ln,Record},RecTuple] ->
	    {attribute,La,record,{Record,record_tuple(RecTuple)}};
	Other -> error_bad_decl(La, record)
    end;
build_attribute({atom,La,file}, Val) ->
    case Val of
	[{string,Ln,Name},{integer,Ll,Line}] ->
	    {attribute,La,file,{Name,Line}};
	Other -> error_bad_decl(La, file)
    end;
build_attribute({atom,La,Attr}, Val) ->
    case Val of
	[Expr] ->
	    {attribute,La,Attr,term(Expr)};
	Other -> return_error(La, "bad attribute")
    end.

error_bad_decl(L, S) ->
    return_error(L, io_lib:format("bad ~w declaration", [S])).

farity_list({cons,Lc,{op,Lo,'/',{atom,La,A},{integer,Li,I}},Tail}) ->
    [{A,I}|farity_list(Tail)];
farity_list({nil,Ln}) -> [];
farity_list(Other) ->
    return_error(line(Other), "bad function arity").

record_tuple({tuple,Lt,Fields}) ->
    record_fields(Fields);
record_tuple(Other) ->
    return_error(line(Other), "bad record declaration").

record_fields([{atom,La,A}|Fields]) ->
    [{record_field,La,{atom,La,A}}|record_fields(Fields)];
record_fields([{match,Lm,{atom,La,A},Expr}|Fields]) ->
    [{record_field,La,{atom,La,A},Expr}|record_fields(Fields)];
record_fields([Other|Fields]) ->
    return_error(line(Other), "bad record field");
record_fields([]) -> [].

term(Expr) ->
    case catch normalise(Expr) of
	{'EXIT',R} -> return_error(line(Expr), "bad attribute");
	Term -> Term
    end.

package_segments(Name) ->
    package_segments(Name, []).

package_segments({record_field,_,F1,F2}, Fs) ->
    package_segments(F1, [F2 | Fs]);
package_segments({atom,La,A}, [F | Fs]) ->
    [A | package_segments(F, Fs)];
package_segments({atom,La,A}, []) ->
    [A];
package_segments(_, _) ->
    error.

%% build_function([Clause]) -> {function,Line,Name,Arity,[Clause]}

build_function(Cs) ->
    Name = element(3, hd(Cs)),
    Arity = length(element(4, hd(Cs))),
    {function,line(hd(Cs)),Name,Arity,check_clauses(Cs, Name, Arity)}.

%% build_rule([Clause]) -> {rule,Line,Name,Arity,[Clause]'}

build_rule(Cs) ->
    Name = element(3, hd(Cs)),
    Arity = length(element(4, hd(Cs))),
    {rule,line(hd(Cs)),Name,Arity,check_clauses(Cs, Name, Arity)}.

%% build_fun(Line, [Clause]) -> {'fun',Line,{clauses,[Clause]}}.

build_fun(Line, Cs) ->
    Arity = length(element(4, hd(Cs))),
    {'fun',Line,{clauses,check_clauses(Cs, 'fun', Arity)}}.

check_clauses(Cs, Name, Arity) ->
     mapl(fun ({clause,L,N,As,G,B}) when N == Name, length(As) == Arity ->
		 {clause,L,As,G,B};
	     ({clause,L,N,As,G,B}) ->
		 return_error(L, "head mismatch") end, Cs).

%% mapl(F,List)
%% an alternative map which always maps from left to right
%% and makes it possible to interrupt the mapping with throw on
%% the first occurence from left as expected.
%% can be removed when the jam machine (and all other machines)
%% uses the standardized (Erlang 5.0) evaluation order (from left to right)
mapl(F, [H|T]) ->
	V = F(H),
	[V | mapl(F,T)];
mapl(_, []) ->
	[].

%% normalise(AbsTerm)
%% abstract(Term)
%%  Convert between the abstract form of a term and a term.

normalise({char,_,C}) -> C;
normalise({integer,_,I}) -> I;
normalise({float,_,F}) -> F;
normalise({atom,_,A}) -> A;
normalise({string,_,S}) -> S;
normalise({nil,_}) -> [];
normalise({bin,_,Fs}) ->
    {value, B, _} =
	eval_bits:expr_grp(Fs, [],
			   fun(E, _) ->
				   {value, normalise(E), []}
			   end, [], true),
    B;
normalise({cons,_,Head,Tail}) ->
    [normalise(Head)|normalise(Tail)];
normalise({tuple,_,Args}) ->
    list_to_tuple(normalise_list(Args));
%% Special case for unary +/-.
normalise({op,_,'+',{char,_,I}}) -> I;
normalise({op,_,'+',{integer,_,I}}) -> I;
normalise({op,_,'+',{float,_,F}}) -> F;
normalise({op,_,'-',{char,_,I}}) -> -I;		%Weird, but compatible!
normalise({op,_,'-',{integer,_,I}}) -> -I;
normalise({op,_,'-',{float,_,F}}) -> -F.

normalise_list([H|T]) ->
    [normalise(H)|normalise_list(T)];
normalise_list([]) ->
    [].

abstract(T) when integer(T) -> {integer,0,T};
abstract(T) when float(T) -> {float,0,T};
abstract(T) when atom(T) -> {atom,0,T};
abstract([]) -> {nil,0};
abstract(B) when binary(B) ->
    {bin, 0, lists:map(fun(Byte) ->
			       {bin_element, 0,
				{integer, 0, Byte}, default, default}
		       end,
		       binary_to_list(B))};
abstract([C|T]) when integer(C), 0 =< C, C < 256 ->
    abstract_string(T, [C]);
abstract([H|T]) ->
    {cons,0,abstract(H),abstract(T)};
abstract(Tuple) when tuple(Tuple) ->
    {tuple,0,abstract_list(tuple_to_list(Tuple))}.

abstract_string([C|T], String) when integer(C), 0 =< C, C < 256 ->
    abstract_string(T, [C|String]);
abstract_string([], String) ->
    {string, 0, lists:reverse(String)};
abstract_string(T, String) ->
    not_string(String, abstract(T)).

not_string([C|T], Result) ->
    not_string(T, {cons, 0, {integer, 0, C}, Result});
not_string([], Result) ->
    Result.

abstract_list([H|T]) ->
    [abstract(H)|abstract_list(T)];
abstract_list([]) ->
    [].

%%% abstract/2 keeps the line number
abstract(T, Line) when integer(T) -> {integer,Line,T};
abstract(T, Line) when float(T) -> {float,Line,T};
abstract(T, Line) when atom(T) -> {atom,Line,T};
abstract([], Line) -> {nil,Line};
abstract(B, Line) when binary(B) ->
    {bin, Line, lists:map(fun(Byte) ->
			       {bin_element, Line,
				{integer, Line, Byte}, default, default}
		       end,
		       binary_to_list(B))};
abstract([C|T], Line) when integer(C), 0 =< C, C < 256 ->
    abstract_string(T, [C], Line);
abstract([H|T], Line) ->
    {cons,Line,abstract(H, Line),abstract(T, Line)};
abstract(Tuple, Line) when tuple(Tuple) ->
    {tuple,Line,abstract_list(tuple_to_list(Tuple), Line)}.

abstract_string([C|T], String, Line) when integer(C), 0 =< C, C < 256 ->
    abstract_string(T, [C|String], Line);
abstract_string([], String, Line) ->
    {string, Line, lists:reverse(String)};
abstract_string(T, String, Line) ->
    not_string(String, abstract(T, Line), Line).

not_string([C|T], Result, Line) ->
    not_string(T, {cons, Line, {integer, Line, C}, Result}, Line);
not_string([], Result, Line) ->
    Result.

abstract_list([H|T], Line) ->
    [abstract(H, Line)|abstract_list(T, Line)];
abstract_list([], Line) ->
    [].

%% tokens(AbsTerm) -> [Token]
%% tokens(AbsTerm, More) -> [Token]
%%  Generate a list of tokens representing the abstract term.

tokens(Abs) ->
    tokens(Abs, []).

tokens({char,L,C}, More) -> [{char,L,C}|More];
tokens({integer,L,N}, More) -> [{integer,L,N}|More];
tokens({float,L,F}, More) -> [{float,L,F}|More];
tokens({atom,L,A}, More) -> [{atom,L,A}|More];
tokens({var,L,V}, More) -> [{var,L,V}|More];
tokens({string,L,S}, More) -> [{string,L,S}|More];
tokens({nil,L}, More) -> [{'[',L},{']',L}|More];
tokens({cons,L,Head,Tail}, More) ->
    [{'[',L}|tokens(Head, tokens_tail(Tail, More))];
tokens({tuple,L,[]}, More) ->
    [{'{',L},{'}',L}|More];
tokens({tuple,L,[E|Es]}, More) ->
    [{'{',L}|tokens(E, tokens_tuple(Es, line(E), More))].

tokens_tail({cons,L,Head,Tail}, More) ->
    [{',',L}|tokens(Head, tokens_tail(Tail, More))];
tokens_tail({nil,L}, More) ->
    [{']',L}|More];
tokens_tail(Other, More) ->
    L = line(Other),
    [{'|',L}|tokens(Other, [{']',L}|More])].

tokens_tuple([E|Es], Line, More) ->
    [{',',Line}|tokens(E, tokens_tuple(Es, line(E), More))];
tokens_tuple([], Line, More) ->
    [{'}',Line}|More].

%% Give the relative precedences of operators.

inop_prec('=') -> {150,100,100};
inop_prec('!') -> {150,100,100};
inop_prec('orelse') -> {160,150,150};
inop_prec('andalso') -> {200,160,160};
inop_prec('==') -> {300,200,300};
inop_prec('/=') -> {300,200,300};
inop_prec('=<') -> {300,200,300};
inop_prec('<') -> {300,200,300};
inop_prec('>=') -> {300,200,300};
inop_prec('>') -> {300,200,300};
inop_prec('=:=') -> {300,200,300};
inop_prec('=/=') -> {300,200,300};
inop_prec('++') -> {300,300,400};
inop_prec('--') -> {300,300,400};
inop_prec('+') -> {400,400,500};
inop_prec('-') -> {400,400,500};
inop_prec('bor') -> {400,400,500};
inop_prec('bxor') -> {400,400,500};
inop_prec('bsl') -> {400,400,500};
inop_prec('bsr') -> {400,400,500};
inop_prec('or') -> {400,400,500};
inop_prec('xor') -> {400,400,500};
inop_prec('*') -> {500,500,600};
inop_prec('/') -> {500,500,600};
inop_prec('div') -> {500,500,600};
inop_prec('rem') -> {500,500,600};
inop_prec('band') -> {500,500,600};
inop_prec('and') -> {500,500,600};
inop_prec('#') -> {800,700,800};
inop_prec(':') -> {900,800,900}.

preop_prec('catch') -> {0,100};
preop_prec('+') -> {600,700};
preop_prec('-') -> {600,700};
preop_prec('bnot') -> {600,700};
preop_prec('not') -> {600,700};
preop_prec('#') -> {700,800}.

func_prec() -> {800,700}.

max_prec() -> 1000.

%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The parser generator will insert appropriate declarations before this line.%

parse(Tokens) ->
    case catch yeccpars1(Tokens, false, 0, [], []) of
	error ->
	    Errorline =
		if Tokens == [] -> 0; true -> element(2, hd(Tokens)) end,
	    {error,
	     {Errorline, ?THIS_MODULE, "syntax error at or after this line."}};
	Other ->
	    Other
    end.

parse_and_scan({Mod, Fun, Args}) ->
    case apply(Mod, Fun, Args) of
	{eof, _} ->
	    {ok, eof};
	{error, Descriptor, _} ->
	    {error, Descriptor};
	{ok, Tokens, _} ->
	    yeccpars1(Tokens, {Mod, Fun, Args}, 0, [], [])
    end.

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
	true ->
	    Message;
	_ ->
	    io_lib:write(Message)
    end.

% To be used in grammar files to throw an error message to the parser toplevel.
% Doesn't have to be exported!
return_error(Line, Message) ->
    throw({error, {Line, ?THIS_MODULE, Message}}).


% Don't change yeccpars1/6 too much, it is called recursively by yeccpars2/8!
yeccpars1([Token | Tokens], Tokenizer, State, States, Vstack) ->
    yeccpars2(State, element(1, Token), States, Vstack, Token, Tokens,
	      Tokenizer);
yeccpars1([], {M, F, A}, State, States, Vstack) ->
    case catch apply(M, F, A) of
        {eof, Endline} ->
            {error, {Endline, ?THIS_MODULE, "end_of_file"}};
        {error, Descriptor, _Endline} ->
            {error, Descriptor};
        {'EXIT', Reason} ->
            {error, {0, ?THIS_MODULE, Reason}};
        {ok, Tokens, _Endline} ->
	    case catch yeccpars1(Tokens, {M, F, A}, State, States, Vstack) of
		error ->
		    Errorline = element(2, hd(Tokens)),
		    {error, {Errorline, ?THIS_MODULE,
			     "syntax error at or after this line."}};
		Other ->
		    Other
	    end
    end;
yeccpars1([], false, State, States, Vstack) ->
    yeccpars2(State, '$end', States, Vstack, {'$end', 999999}, [], false).

% For internal use only.
yeccerror(Token) ->
    {error,
     {element(2, Token), ?THIS_MODULE,
      ["syntax error before: ", yecctoken2string(Token)]}}.

yecctoken2string({atom, _, A}) -> io_lib:write(A);
yecctoken2string({integer,_,N}) -> io_lib:write(N);
yecctoken2string({float,_,F}) -> io_lib:write(F);
yecctoken2string({char,_,C}) -> io_lib:write_char(C);
yecctoken2string({var,_,V}) -> io_lib:format('~s', [V]);
yecctoken2string({string,_,S}) -> io_lib:write_string(S);
yecctoken2string({reserved_symbol, _, A}) -> io_lib:format('~w', [A]);
yecctoken2string({_Cat, _, Val}) -> io_lib:format('~w', [Val]);

yecctoken2string({'dot', _}) -> io_lib:format('~w', ['.']);
yecctoken2string({'$end', _}) ->
    [];
yecctoken2string({Other, _}) when atom(Other) ->
    io_lib:format('~w', [Other]);
yecctoken2string(Other) ->
    io_lib:write(Other).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


yeccpars2(0, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 2, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(1, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 269, [1 | __Ss], [__T | __Stack]);
yeccpars2(1, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(2, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [2 | __Ss], [__T | __Stack]);
yeccpars2(2, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(3, 'dot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 266, [3 | __Ss], [__T | __Stack]);
yeccpars2(3, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(4, '$end', _, __Stack, _, _, _) ->
 {ok, hd(__Stack)};
yeccpars2(4, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(5, 'dot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 265, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(6, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 259, [6 | __Ss], [__T | __Stack]);
yeccpars2(6, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(function_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(7, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  build_function(__1),
 yeccpars2(yeccgoto(function, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(8, 'dot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 258, [8 | __Ss], [__T | __Stack]);
yeccpars2(8, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(9, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [9 | __Ss], [__T | __Stack]);
yeccpars2(9, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(rule_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(10, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  build_rule(__1),
 yeccpars2(yeccgoto(rule, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(11, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [11 | __Ss], [__T | __Stack]);
yeccpars2(11, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(12, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [12 | __Ss], [__T | __Stack]);
yeccpars2(12, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(13, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(rule_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(14, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 255, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(15, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto(clause_args, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(16, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [16 | __Ss], [__T | __Stack]);
yeccpars2(16, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(17, __Cat, [16 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(17, ':-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 252, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(18, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(19, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 248, [19 | __Ss], [__T | __Stack]);
yeccpars2(19, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(20, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(21, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(prefix_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(22, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(prefix_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(23, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 245, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(24, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '>>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 222, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(25, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 212, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(26, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(27, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(28, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(29, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(30, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(prefix_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(31, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(32, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(33, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(34, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(35, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 203, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(exprs, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(36, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(37, '!', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 199, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 200, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_100, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(38, 'orelse', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_150, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(39, 'andalso', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 195, [39 | __Ss], [__T | __Stack]);
yeccpars2(39, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_160, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(40, '=/=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 187, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '=:=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 188, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 191, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '>=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 192, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 186, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '=<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 189, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '/=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 185, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '==', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 190, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_200, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(41, '--', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 174, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, '++', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 172, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'xor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 182, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'or', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 181, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'bsr', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 178, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'bsl', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 177, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'bxor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 179, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'bor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 176, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 173, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 171, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_300, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(42, 'and', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [42 | __Ss], [__T | __Stack]);
yeccpars2(42, 'band', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [42 | __Ss], [__T | __Stack]);
yeccpars2(42, 'rem', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 169, [42 | __Ss], [__T | __Stack]);
yeccpars2(42, 'div', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 167, [42 | __Ss], [__T | __Stack]);
yeccpars2(42, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [42 | __Ss], [__T | __Stack]);
yeccpars2(42, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [42 | __Ss], [__T | __Stack]);
yeccpars2(42, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_400, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(43, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_500, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(44, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_600, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(45, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [45 | __Ss], [__T | __Stack]);
yeccpars2(45, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_700, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(46, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 158, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 159, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_800, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(47, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_900, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(48, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(guard, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(49, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(50, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [50 | __Ss], [__T | __Stack]);
yeccpars2(50, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 128, [50 | __Ss], [__T | __Stack]);
yeccpars2(50, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(51, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(52, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_700, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(53, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(clause_guard, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(54, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(55, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(56, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(57, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(58, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(59, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(prefix_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(60, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(61, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(62, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(63, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 92, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'after', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(64, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(65, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_700, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(66, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(strings, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(67, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(68, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [68 | __Ss], [__T | __Stack]);
yeccpars2(68, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(69, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(70, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(71, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(72, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [72 | __Ss], [__T | __Stack]);
yeccpars2(72, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(73, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [73 | __Ss], [__T | __Stack]);
yeccpars2(73, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(74, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {tuple,line(__1),[]},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(75, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {tuple,line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(76, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [76 | __Ss], [__T | __Stack]);
yeccpars2(76, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [76 | __Ss], [__T | __Stack]);
yeccpars2(76, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(77, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(78, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'try',line(__1),__2,[]},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(try_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(79, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(cr_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(80, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(81, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(82, __Cat, [81 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(82, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(83, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(84, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {clause,line(__1),[__1],__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(cr_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(85, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(clause_body, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(86, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'try',line(__1),__2,__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(try_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(87, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(88, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(cr_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(89, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {string,line(__1),element(3,__1) ++ element(3,__2)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(strings, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(90, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(91, 'after', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 99, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(92, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'after', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(93, 'after', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(94, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(95, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'receive',line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(96, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(97, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(98, '->', __Ss, [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'receive',line(__1),__2,__4,__5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), '->', __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(98, 'when', __Ss, [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'receive',line(__1),__2,__4,__5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), 'when', __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(98, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'receive',line(__1),__2,__4,__5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(99, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(100, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(101, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(102, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'receive',line(__1),__2,__4,__5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(103, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [103 | __Ss], [__T | __Stack]);
yeccpars2(103, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(104, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 105, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(105, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'receive',line(__1),[],__3,__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(106, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(107, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 108, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(108, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'query',line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(query_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(109, '||', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 110, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(110, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(111, '<-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 117, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(lc_expr, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(112, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [112 | __Ss], [__T | __Stack]);
yeccpars2(112, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(lc_exprs, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(113, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(114, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {lc,line(__1),__2,__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(list_comprehension, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(115, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(116, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(lc_exprs, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(117, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(118, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {generate,line(__2),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(lc_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(119, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(expr_600, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(120, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(121, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [121 | __Ss], [__T | __Stack]);
yeccpars2(121, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(if_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(122, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [122 | __Ss], [__T | __Stack]);
yeccpars2(122, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(123, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'if',line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(if_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(124, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(125, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(if_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(126, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {clause,line(hd(hd(__1))),[],__1,__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(if_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(127, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(136, __Cat, [127 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(128, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [128 | __Ss], [__T | __Stack]);
yeccpars2(128, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(129, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 132, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(fun_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(130, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(131, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  build_fun(line(__1),__2),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(fun_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(132, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(133, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(fun_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(134, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(135, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'fun',line(__1),{function,element(3,__2),element(3,__4)}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(fun_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(136, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [136 | __Ss], [__T | __Stack]);
yeccpars2(136, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(137, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {Args,Pos} = __1, {clause,Pos,'fun',Args,__2,__3}
  end,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(fun_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(138, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(139, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(guard, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(140, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(141, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(142, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 157, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(143, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record,line(__2),__1,element(3,__3),__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(record_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(144, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 149, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(145, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 155, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(146, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(record_fields, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(147, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 152, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(148, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 150, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(149, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(record_tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(150, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(151, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_field,line(__1),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_field, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(152, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(153, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [153 | __Ss], [__T | __Stack]);
yeccpars2(153, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [153 | __Ss], [__T | __Stack]);
yeccpars2(153, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(154, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_fields, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(155, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(156, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_field,line(__1),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_field, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(157, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_field,line(__2),__1,element(3,__3),__5},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(record_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(158, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 161, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(159, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(160, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {remote,line(__2),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_800, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(161, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_field,line(__2),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_900, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(162, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {call,line(__1),__1,element(1,__2)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(function_call, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(163, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(164, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(165, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(166, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(167, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(168, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(169, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(170, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_500, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(171, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(172, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(list_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(173, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(174, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(list_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(175, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(176, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(177, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(178, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(179, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(180, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(181, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(182, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(183, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_300, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(184, 'and', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'band', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'rem', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 169, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'div', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 167, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_400, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(185, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(186, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(187, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(188, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(189, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(190, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(191, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(192, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(193, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(194, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_200, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(195, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(196, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_160, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(197, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(198, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_150, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(199, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(200, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(201, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {match,line(__2),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_100, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(202, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_100, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(203, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(204, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(exprs, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(205, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'catch',line(__1),__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(206, 'of', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 207, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(207, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(208, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 209, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(209, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'case',line(__1),__2,__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(case_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(210, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 211, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(211, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {block,line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_max, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(212, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {nil,line(__1)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(213, '||', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 110, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 214, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 217, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(214, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(215, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {nil,line(__1)},
 yeccpars2(yeccgoto(tail, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(216, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {cons,line(__1),__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(217, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(218, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 219, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(219, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(220, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 214, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 217, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(221, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {cons,line(__2),__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(222, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {bin,line(__1),[]},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(binary, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(223, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 243, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(bin_elements, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(224, '>>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 242, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(225, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 229, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  default,
 yeccpars2(230, __Cat, [225 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(226, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(bit_expr, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(227, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(228, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(bit_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(229, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(230, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 231, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  default,
 yeccpars2(232, __Cat, [230 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(231, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 233, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(232, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {bin_element,line(__1),__1,__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(bin_element, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(233, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 238, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto(bit_type, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(234, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 236, [234 | __Ss], [__T | __Stack]);
yeccpars2(234, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(bit_type_list, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(235, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(opt_bit_type_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(236, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 233, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(237, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(bit_type_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(238, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 239, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(239, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {element(3,__1),element(3,__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(bit_type, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(240, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(opt_bit_size_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(241, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(bit_size_expr, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(242, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {bin,line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(binary, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(243, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(244, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(bin_elements, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(245, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_field,line(__1),{atom,line(__1),''},__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(expr_900, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(246, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 247, [246 | __Ss], [__T | __Stack]);
yeccpars2(246, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(247, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_max, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(248, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [248 | __Ss], [__T | __Stack]);
yeccpars2(248, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 249, [248 | __Ss], [__T | __Stack]);
yeccpars2(248, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(249, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 251, [249 | __Ss], [__T | __Stack]);
yeccpars2(249, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(250, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record,line(__1),element(3,__2),__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(251, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_index,line(__1),element(3,__2),__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(record_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(252, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(253, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {clause,line(__1),element(3,__1),__2,__3,__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(rule_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(254, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(rule_body, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(255, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {[],line(__1)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(argument_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(256, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 257, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(257, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__2,line(__1)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(argument_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(258, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(form, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(259, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 260, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(260, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(261, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(function_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(262, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [262 | __Ss], [__T | __Stack]);
yeccpars2(262, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(263, __Cat, [262 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(263, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [263 | __Ss], [__T | __Stack]);
yeccpars2(263, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(264, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {clause,line(__1),element(3,__1),__2,__3,__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(function_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(265, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(form, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(266, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(form, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(267, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(268, __Cat, [267 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(268, ':-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 252, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(269, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 270, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(270, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(271, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 273, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(272, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(attr_val, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(273, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  build_attribute(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(attribute, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(__Other, _, _, _, _, _, _) ->
 exit({parser, __Other, missing_state_in_action_table}).

yeccgoto(add_op, 41) ->
 175;
yeccgoto(argument_list, 2) ->
 15;
yeccgoto(argument_list, 12) ->
 15;
yeccgoto(argument_list, 45) ->
 162;
yeccgoto(argument_list, 50) ->
 127;
yeccgoto(argument_list, 132) ->
 127;
yeccgoto(argument_list, 260) ->
 15;
yeccgoto(atomic, 14) ->
 27;
yeccgoto(atomic, 18) ->
 27;
yeccgoto(atomic, 20) ->
 27;
yeccgoto(atomic, 24) ->
 27;
yeccgoto(atomic, 25) ->
 27;
yeccgoto(atomic, 28) ->
 27;
yeccgoto(atomic, 31) ->
 27;
yeccgoto(atomic, 33) ->
 27;
yeccgoto(atomic, 54) ->
 27;
yeccgoto(atomic, 60) ->
 27;
yeccgoto(atomic, 63) ->
 27;
yeccgoto(atomic, 68) ->
 27;
yeccgoto(atomic, 72) ->
 27;
yeccgoto(atomic, 77) ->
 27;
yeccgoto(atomic, 83) ->
 27;
yeccgoto(atomic, 87) ->
 27;
yeccgoto(atomic, 90) ->
 27;
yeccgoto(atomic, 92) ->
 27;
yeccgoto(atomic, 94) ->
 27;
yeccgoto(atomic, 99) ->
 27;
yeccgoto(atomic, 106) ->
 27;
yeccgoto(atomic, 110) ->
 27;
yeccgoto(atomic, 115) ->
 27;
yeccgoto(atomic, 117) ->
 27;
yeccgoto(atomic, 124) ->
 27;
yeccgoto(atomic, 138) ->
 27;
yeccgoto(atomic, 150) ->
 27;
yeccgoto(atomic, 155) ->
 27;
yeccgoto(atomic, 159) ->
 27;
yeccgoto(atomic, 168) ->
 27;
yeccgoto(atomic, 175) ->
 27;
yeccgoto(atomic, 180) ->
 27;
yeccgoto(atomic, 193) ->
 27;
yeccgoto(atomic, 195) ->
 27;
yeccgoto(atomic, 197) ->
 27;
yeccgoto(atomic, 199) ->
 27;
yeccgoto(atomic, 200) ->
 27;
yeccgoto(atomic, 203) ->
 27;
yeccgoto(atomic, 207) ->
 27;
yeccgoto(atomic, 214) ->
 27;
yeccgoto(atomic, 217) ->
 27;
yeccgoto(atomic, 227) ->
 27;
yeccgoto(atomic, 229) ->
 27;
yeccgoto(atomic, 243) ->
 27;
yeccgoto(atomic, 252) ->
 27;
yeccgoto(atomic, 270) ->
 27;
yeccgoto(attr_val, 270) ->
 271;
yeccgoto(attribute, 0) ->
 3;
yeccgoto(bin_element, 24) ->
 223;
yeccgoto(bin_element, 243) ->
 223;
yeccgoto(bin_elements, 24) ->
 224;
yeccgoto(bin_elements, 243) ->
 244;
yeccgoto(binary, 14) ->
 29;
yeccgoto(binary, 18) ->
 29;
yeccgoto(binary, 20) ->
 29;
yeccgoto(binary, 24) ->
 29;
yeccgoto(binary, 25) ->
 29;
yeccgoto(binary, 28) ->
 29;
yeccgoto(binary, 31) ->
 29;
yeccgoto(binary, 33) ->
 29;
yeccgoto(binary, 54) ->
 29;
yeccgoto(binary, 60) ->
 29;
yeccgoto(binary, 63) ->
 29;
yeccgoto(binary, 68) ->
 29;
yeccgoto(binary, 72) ->
 29;
yeccgoto(binary, 77) ->
 29;
yeccgoto(binary, 83) ->
 29;
yeccgoto(binary, 87) ->
 29;
yeccgoto(binary, 90) ->
 29;
yeccgoto(binary, 92) ->
 29;
yeccgoto(binary, 94) ->
 29;
yeccgoto(binary, 99) ->
 29;
yeccgoto(binary, 106) ->
 29;
yeccgoto(binary, 110) ->
 29;
yeccgoto(binary, 115) ->
 29;
yeccgoto(binary, 117) ->
 29;
yeccgoto(binary, 124) ->
 29;
yeccgoto(binary, 138) ->
 29;
yeccgoto(binary, 150) ->
 29;
yeccgoto(binary, 155) ->
 29;
yeccgoto(binary, 159) ->
 29;
yeccgoto(binary, 168) ->
 29;
yeccgoto(binary, 175) ->
 29;
yeccgoto(binary, 180) ->
 29;
yeccgoto(binary, 193) ->
 29;
yeccgoto(binary, 195) ->
 29;
yeccgoto(binary, 197) ->
 29;
yeccgoto(binary, 199) ->
 29;
yeccgoto(binary, 200) ->
 29;
yeccgoto(binary, 203) ->
 29;
yeccgoto(binary, 207) ->
 29;
yeccgoto(binary, 214) ->
 29;
yeccgoto(binary, 217) ->
 29;
yeccgoto(binary, 227) ->
 29;
yeccgoto(binary, 229) ->
 29;
yeccgoto(binary, 243) ->
 29;
yeccgoto(binary, 252) ->
 29;
yeccgoto(binary, 270) ->
 29;
yeccgoto(bit_expr, 24) ->
 225;
yeccgoto(bit_expr, 243) ->
 225;
yeccgoto(bit_size_expr, 229) ->
 240;
yeccgoto(bit_type, 231) ->
 234;
yeccgoto(bit_type, 236) ->
 234;
yeccgoto(bit_type_list, 231) ->
 235;
yeccgoto(bit_type_list, 236) ->
 237;
yeccgoto(case_expr, 14) ->
 32;
yeccgoto(case_expr, 18) ->
 32;
yeccgoto(case_expr, 20) ->
 32;
yeccgoto(case_expr, 24) ->
 32;
yeccgoto(case_expr, 25) ->
 32;
yeccgoto(case_expr, 28) ->
 32;
yeccgoto(case_expr, 31) ->
 32;
yeccgoto(case_expr, 33) ->
 32;
yeccgoto(case_expr, 54) ->
 32;
yeccgoto(case_expr, 60) ->
 32;
yeccgoto(case_expr, 63) ->
 32;
yeccgoto(case_expr, 68) ->
 32;
yeccgoto(case_expr, 72) ->
 32;
yeccgoto(case_expr, 77) ->
 32;
yeccgoto(case_expr, 83) ->
 32;
yeccgoto(case_expr, 87) ->
 32;
yeccgoto(case_expr, 90) ->
 32;
yeccgoto(case_expr, 92) ->
 32;
yeccgoto(case_expr, 94) ->
 32;
yeccgoto(case_expr, 99) ->
 32;
yeccgoto(case_expr, 106) ->
 32;
yeccgoto(case_expr, 110) ->
 32;
yeccgoto(case_expr, 115) ->
 32;
yeccgoto(case_expr, 117) ->
 32;
yeccgoto(case_expr, 124) ->
 32;
yeccgoto(case_expr, 138) ->
 32;
yeccgoto(case_expr, 150) ->
 32;
yeccgoto(case_expr, 155) ->
 32;
yeccgoto(case_expr, 159) ->
 32;
yeccgoto(case_expr, 168) ->
 32;
yeccgoto(case_expr, 175) ->
 32;
yeccgoto(case_expr, 180) ->
 32;
yeccgoto(case_expr, 193) ->
 32;
yeccgoto(case_expr, 195) ->
 32;
yeccgoto(case_expr, 197) ->
 32;
yeccgoto(case_expr, 199) ->
 32;
yeccgoto(case_expr, 200) ->
 32;
yeccgoto(case_expr, 203) ->
 32;
yeccgoto(case_expr, 207) ->
 32;
yeccgoto(case_expr, 214) ->
 32;
yeccgoto(case_expr, 217) ->
 32;
yeccgoto(case_expr, 227) ->
 32;
yeccgoto(case_expr, 229) ->
 32;
yeccgoto(case_expr, 243) ->
 32;
yeccgoto(case_expr, 252) ->
 32;
yeccgoto(case_expr, 270) ->
 32;
yeccgoto(clause_args, 2) ->
 267;
yeccgoto(clause_args, 12) ->
 16;
yeccgoto(clause_args, 260) ->
 262;
yeccgoto(clause_body, 82) ->
 84;
yeccgoto(clause_body, 96) ->
 97;
yeccgoto(clause_body, 100) ->
 101;
yeccgoto(clause_body, 103) ->
 104;
yeccgoto(clause_body, 120) ->
 126;
yeccgoto(clause_body, 136) ->
 137;
yeccgoto(clause_body, 263) ->
 264;
yeccgoto(clause_body, 268) ->
 264;
yeccgoto(clause_guard, 16) ->
 17;
yeccgoto(clause_guard, 81) ->
 82;
yeccgoto(clause_guard, 127) ->
 136;
yeccgoto(clause_guard, 262) ->
 263;
yeccgoto(clause_guard, 267) ->
 268;
yeccgoto(comp_op, 40) ->
 193;
yeccgoto(cr_clause, 63) ->
 79;
yeccgoto(cr_clause, 77) ->
 79;
yeccgoto(cr_clause, 87) ->
 79;
yeccgoto(cr_clause, 92) ->
 79;
yeccgoto(cr_clause, 207) ->
 79;
yeccgoto(cr_clauses, 63) ->
 91;
yeccgoto(cr_clauses, 77) ->
 80;
yeccgoto(cr_clauses, 87) ->
 88;
yeccgoto(cr_clauses, 92) ->
 93;
yeccgoto(cr_clauses, 207) ->
 208;
yeccgoto(expr, 14) ->
 35;
yeccgoto(expr, 18) ->
 35;
yeccgoto(expr, 20) ->
 246;
yeccgoto(expr, 25) ->
 213;
yeccgoto(expr, 28) ->
 35;
yeccgoto(expr, 31) ->
 206;
yeccgoto(expr, 33) ->
 205;
yeccgoto(expr, 54) ->
 35;
yeccgoto(expr, 63) ->
 81;
yeccgoto(expr, 68) ->
 35;
yeccgoto(expr, 72) ->
 35;
yeccgoto(expr, 77) ->
 81;
yeccgoto(expr, 83) ->
 35;
yeccgoto(expr, 87) ->
 81;
yeccgoto(expr, 90) ->
 103;
yeccgoto(expr, 92) ->
 81;
yeccgoto(expr, 94) ->
 96;
yeccgoto(expr, 99) ->
 100;
yeccgoto(expr, 106) ->
 109;
yeccgoto(expr, 110) ->
 111;
yeccgoto(expr, 115) ->
 111;
yeccgoto(expr, 117) ->
 118;
yeccgoto(expr, 124) ->
 35;
yeccgoto(expr, 138) ->
 35;
yeccgoto(expr, 150) ->
 151;
yeccgoto(expr, 155) ->
 156;
yeccgoto(expr, 203) ->
 35;
yeccgoto(expr, 207) ->
 81;
yeccgoto(expr, 214) ->
 220;
yeccgoto(expr, 217) ->
 218;
yeccgoto(expr, 252) ->
 111;
yeccgoto(expr, 270) ->
 35;
yeccgoto(expr_100, 14) ->
 36;
yeccgoto(expr_100, 18) ->
 36;
yeccgoto(expr_100, 20) ->
 36;
yeccgoto(expr_100, 25) ->
 36;
yeccgoto(expr_100, 28) ->
 36;
yeccgoto(expr_100, 31) ->
 36;
yeccgoto(expr_100, 33) ->
 36;
yeccgoto(expr_100, 54) ->
 36;
yeccgoto(expr_100, 63) ->
 36;
yeccgoto(expr_100, 68) ->
 36;
yeccgoto(expr_100, 72) ->
 36;
yeccgoto(expr_100, 77) ->
 36;
yeccgoto(expr_100, 83) ->
 36;
yeccgoto(expr_100, 87) ->
 36;
yeccgoto(expr_100, 90) ->
 36;
yeccgoto(expr_100, 92) ->
 36;
yeccgoto(expr_100, 94) ->
 36;
yeccgoto(expr_100, 99) ->
 36;
yeccgoto(expr_100, 106) ->
 36;
yeccgoto(expr_100, 110) ->
 36;
yeccgoto(expr_100, 115) ->
 36;
yeccgoto(expr_100, 117) ->
 36;
yeccgoto(expr_100, 124) ->
 36;
yeccgoto(expr_100, 138) ->
 36;
yeccgoto(expr_100, 150) ->
 36;
yeccgoto(expr_100, 155) ->
 36;
yeccgoto(expr_100, 199) ->
 202;
yeccgoto(expr_100, 200) ->
 201;
yeccgoto(expr_100, 203) ->
 36;
yeccgoto(expr_100, 207) ->
 36;
yeccgoto(expr_100, 214) ->
 36;
yeccgoto(expr_100, 217) ->
 36;
yeccgoto(expr_100, 252) ->
 36;
yeccgoto(expr_100, 270) ->
 36;
yeccgoto(expr_150, 14) ->
 37;
yeccgoto(expr_150, 18) ->
 37;
yeccgoto(expr_150, 20) ->
 37;
yeccgoto(expr_150, 25) ->
 37;
yeccgoto(expr_150, 28) ->
 37;
yeccgoto(expr_150, 31) ->
 37;
yeccgoto(expr_150, 33) ->
 37;
yeccgoto(expr_150, 54) ->
 37;
yeccgoto(expr_150, 63) ->
 37;
yeccgoto(expr_150, 68) ->
 37;
yeccgoto(expr_150, 72) ->
 37;
yeccgoto(expr_150, 77) ->
 37;
yeccgoto(expr_150, 83) ->
 37;
yeccgoto(expr_150, 87) ->
 37;
yeccgoto(expr_150, 90) ->
 37;
yeccgoto(expr_150, 92) ->
 37;
yeccgoto(expr_150, 94) ->
 37;
yeccgoto(expr_150, 99) ->
 37;
yeccgoto(expr_150, 106) ->
 37;
yeccgoto(expr_150, 110) ->
 37;
yeccgoto(expr_150, 115) ->
 37;
yeccgoto(expr_150, 117) ->
 37;
yeccgoto(expr_150, 124) ->
 37;
yeccgoto(expr_150, 138) ->
 37;
yeccgoto(expr_150, 150) ->
 37;
yeccgoto(expr_150, 155) ->
 37;
yeccgoto(expr_150, 197) ->
 198;
yeccgoto(expr_150, 199) ->
 37;
yeccgoto(expr_150, 200) ->
 37;
yeccgoto(expr_150, 203) ->
 37;
yeccgoto(expr_150, 207) ->
 37;
yeccgoto(expr_150, 214) ->
 37;
yeccgoto(expr_150, 217) ->
 37;
yeccgoto(expr_150, 252) ->
 37;
yeccgoto(expr_150, 270) ->
 37;
yeccgoto(expr_160, 14) ->
 38;
yeccgoto(expr_160, 18) ->
 38;
yeccgoto(expr_160, 20) ->
 38;
yeccgoto(expr_160, 25) ->
 38;
yeccgoto(expr_160, 28) ->
 38;
yeccgoto(expr_160, 31) ->
 38;
yeccgoto(expr_160, 33) ->
 38;
yeccgoto(expr_160, 54) ->
 38;
yeccgoto(expr_160, 63) ->
 38;
yeccgoto(expr_160, 68) ->
 38;
yeccgoto(expr_160, 72) ->
 38;
yeccgoto(expr_160, 77) ->
 38;
yeccgoto(expr_160, 83) ->
 38;
yeccgoto(expr_160, 87) ->
 38;
yeccgoto(expr_160, 90) ->
 38;
yeccgoto(expr_160, 92) ->
 38;
yeccgoto(expr_160, 94) ->
 38;
yeccgoto(expr_160, 99) ->
 38;
yeccgoto(expr_160, 106) ->
 38;
yeccgoto(expr_160, 110) ->
 38;
yeccgoto(expr_160, 115) ->
 38;
yeccgoto(expr_160, 117) ->
 38;
yeccgoto(expr_160, 124) ->
 38;
yeccgoto(expr_160, 138) ->
 38;
yeccgoto(expr_160, 150) ->
 38;
yeccgoto(expr_160, 155) ->
 38;
yeccgoto(expr_160, 195) ->
 196;
yeccgoto(expr_160, 197) ->
 38;
yeccgoto(expr_160, 199) ->
 38;
yeccgoto(expr_160, 200) ->
 38;
yeccgoto(expr_160, 203) ->
 38;
yeccgoto(expr_160, 207) ->
 38;
yeccgoto(expr_160, 214) ->
 38;
yeccgoto(expr_160, 217) ->
 38;
yeccgoto(expr_160, 252) ->
 38;
yeccgoto(expr_160, 270) ->
 38;
yeccgoto(expr_200, 14) ->
 39;
yeccgoto(expr_200, 18) ->
 39;
yeccgoto(expr_200, 20) ->
 39;
yeccgoto(expr_200, 25) ->
 39;
yeccgoto(expr_200, 28) ->
 39;
yeccgoto(expr_200, 31) ->
 39;
yeccgoto(expr_200, 33) ->
 39;
yeccgoto(expr_200, 54) ->
 39;
yeccgoto(expr_200, 63) ->
 39;
yeccgoto(expr_200, 68) ->
 39;
yeccgoto(expr_200, 72) ->
 39;
yeccgoto(expr_200, 77) ->
 39;
yeccgoto(expr_200, 83) ->
 39;
yeccgoto(expr_200, 87) ->
 39;
yeccgoto(expr_200, 90) ->
 39;
yeccgoto(expr_200, 92) ->
 39;
yeccgoto(expr_200, 94) ->
 39;
yeccgoto(expr_200, 99) ->
 39;
yeccgoto(expr_200, 106) ->
 39;
yeccgoto(expr_200, 110) ->
 39;
yeccgoto(expr_200, 115) ->
 39;
yeccgoto(expr_200, 117) ->
 39;
yeccgoto(expr_200, 124) ->
 39;
yeccgoto(expr_200, 138) ->
 39;
yeccgoto(expr_200, 150) ->
 39;
yeccgoto(expr_200, 155) ->
 39;
yeccgoto(expr_200, 195) ->
 39;
yeccgoto(expr_200, 197) ->
 39;
yeccgoto(expr_200, 199) ->
 39;
yeccgoto(expr_200, 200) ->
 39;
yeccgoto(expr_200, 203) ->
 39;
yeccgoto(expr_200, 207) ->
 39;
yeccgoto(expr_200, 214) ->
 39;
yeccgoto(expr_200, 217) ->
 39;
yeccgoto(expr_200, 252) ->
 39;
yeccgoto(expr_200, 270) ->
 39;
yeccgoto(expr_300, 14) ->
 40;
yeccgoto(expr_300, 18) ->
 40;
yeccgoto(expr_300, 20) ->
 40;
yeccgoto(expr_300, 25) ->
 40;
yeccgoto(expr_300, 28) ->
 40;
yeccgoto(expr_300, 31) ->
 40;
yeccgoto(expr_300, 33) ->
 40;
yeccgoto(expr_300, 54) ->
 40;
yeccgoto(expr_300, 63) ->
 40;
yeccgoto(expr_300, 68) ->
 40;
yeccgoto(expr_300, 72) ->
 40;
yeccgoto(expr_300, 77) ->
 40;
yeccgoto(expr_300, 83) ->
 40;
yeccgoto(expr_300, 87) ->
 40;
yeccgoto(expr_300, 90) ->
 40;
yeccgoto(expr_300, 92) ->
 40;
yeccgoto(expr_300, 94) ->
 40;
yeccgoto(expr_300, 99) ->
 40;
yeccgoto(expr_300, 106) ->
 40;
yeccgoto(expr_300, 110) ->
 40;
yeccgoto(expr_300, 115) ->
 40;
yeccgoto(expr_300, 117) ->
 40;
yeccgoto(expr_300, 124) ->
 40;
yeccgoto(expr_300, 138) ->
 40;
yeccgoto(expr_300, 150) ->
 40;
yeccgoto(expr_300, 155) ->
 40;
yeccgoto(expr_300, 180) ->
 183;
yeccgoto(expr_300, 193) ->
 194;
yeccgoto(expr_300, 195) ->
 40;
yeccgoto(expr_300, 197) ->
 40;
yeccgoto(expr_300, 199) ->
 40;
yeccgoto(expr_300, 200) ->
 40;
yeccgoto(expr_300, 203) ->
 40;
yeccgoto(expr_300, 207) ->
 40;
yeccgoto(expr_300, 214) ->
 40;
yeccgoto(expr_300, 217) ->
 40;
yeccgoto(expr_300, 252) ->
 40;
yeccgoto(expr_300, 270) ->
 40;
yeccgoto(expr_400, 14) ->
 41;
yeccgoto(expr_400, 18) ->
 41;
yeccgoto(expr_400, 20) ->
 41;
yeccgoto(expr_400, 25) ->
 41;
yeccgoto(expr_400, 28) ->
 41;
yeccgoto(expr_400, 31) ->
 41;
yeccgoto(expr_400, 33) ->
 41;
yeccgoto(expr_400, 54) ->
 41;
yeccgoto(expr_400, 63) ->
 41;
yeccgoto(expr_400, 68) ->
 41;
yeccgoto(expr_400, 72) ->
 41;
yeccgoto(expr_400, 77) ->
 41;
yeccgoto(expr_400, 83) ->
 41;
yeccgoto(expr_400, 87) ->
 41;
yeccgoto(expr_400, 90) ->
 41;
yeccgoto(expr_400, 92) ->
 41;
yeccgoto(expr_400, 94) ->
 41;
yeccgoto(expr_400, 99) ->
 41;
yeccgoto(expr_400, 106) ->
 41;
yeccgoto(expr_400, 110) ->
 41;
yeccgoto(expr_400, 115) ->
 41;
yeccgoto(expr_400, 117) ->
 41;
yeccgoto(expr_400, 124) ->
 41;
yeccgoto(expr_400, 138) ->
 41;
yeccgoto(expr_400, 150) ->
 41;
yeccgoto(expr_400, 155) ->
 41;
yeccgoto(expr_400, 180) ->
 41;
yeccgoto(expr_400, 193) ->
 41;
yeccgoto(expr_400, 195) ->
 41;
yeccgoto(expr_400, 197) ->
 41;
yeccgoto(expr_400, 199) ->
 41;
yeccgoto(expr_400, 200) ->
 41;
yeccgoto(expr_400, 203) ->
 41;
yeccgoto(expr_400, 207) ->
 41;
yeccgoto(expr_400, 214) ->
 41;
yeccgoto(expr_400, 217) ->
 41;
yeccgoto(expr_400, 252) ->
 41;
yeccgoto(expr_400, 270) ->
 41;
yeccgoto(expr_500, 14) ->
 42;
yeccgoto(expr_500, 18) ->
 42;
yeccgoto(expr_500, 20) ->
 42;
yeccgoto(expr_500, 25) ->
 42;
yeccgoto(expr_500, 28) ->
 42;
yeccgoto(expr_500, 31) ->
 42;
yeccgoto(expr_500, 33) ->
 42;
yeccgoto(expr_500, 54) ->
 42;
yeccgoto(expr_500, 63) ->
 42;
yeccgoto(expr_500, 68) ->
 42;
yeccgoto(expr_500, 72) ->
 42;
yeccgoto(expr_500, 77) ->
 42;
yeccgoto(expr_500, 83) ->
 42;
yeccgoto(expr_500, 87) ->
 42;
yeccgoto(expr_500, 90) ->
 42;
yeccgoto(expr_500, 92) ->
 42;
yeccgoto(expr_500, 94) ->
 42;
yeccgoto(expr_500, 99) ->
 42;
yeccgoto(expr_500, 106) ->
 42;
yeccgoto(expr_500, 110) ->
 42;
yeccgoto(expr_500, 115) ->
 42;
yeccgoto(expr_500, 117) ->
 42;
yeccgoto(expr_500, 124) ->
 42;
yeccgoto(expr_500, 138) ->
 42;
yeccgoto(expr_500, 150) ->
 42;
yeccgoto(expr_500, 155) ->
 42;
yeccgoto(expr_500, 175) ->
 184;
yeccgoto(expr_500, 180) ->
 42;
yeccgoto(expr_500, 193) ->
 42;
yeccgoto(expr_500, 195) ->
 42;
yeccgoto(expr_500, 197) ->
 42;
yeccgoto(expr_500, 199) ->
 42;
yeccgoto(expr_500, 200) ->
 42;
yeccgoto(expr_500, 203) ->
 42;
yeccgoto(expr_500, 207) ->
 42;
yeccgoto(expr_500, 214) ->
 42;
yeccgoto(expr_500, 217) ->
 42;
yeccgoto(expr_500, 252) ->
 42;
yeccgoto(expr_500, 270) ->
 42;
yeccgoto(expr_600, 14) ->
 43;
yeccgoto(expr_600, 18) ->
 43;
yeccgoto(expr_600, 20) ->
 43;
yeccgoto(expr_600, 25) ->
 43;
yeccgoto(expr_600, 28) ->
 43;
yeccgoto(expr_600, 31) ->
 43;
yeccgoto(expr_600, 33) ->
 43;
yeccgoto(expr_600, 54) ->
 43;
yeccgoto(expr_600, 63) ->
 43;
yeccgoto(expr_600, 68) ->
 43;
yeccgoto(expr_600, 72) ->
 43;
yeccgoto(expr_600, 77) ->
 43;
yeccgoto(expr_600, 83) ->
 43;
yeccgoto(expr_600, 87) ->
 43;
yeccgoto(expr_600, 90) ->
 43;
yeccgoto(expr_600, 92) ->
 43;
yeccgoto(expr_600, 94) ->
 43;
yeccgoto(expr_600, 99) ->
 43;
yeccgoto(expr_600, 106) ->
 43;
yeccgoto(expr_600, 110) ->
 43;
yeccgoto(expr_600, 115) ->
 43;
yeccgoto(expr_600, 117) ->
 43;
yeccgoto(expr_600, 124) ->
 43;
yeccgoto(expr_600, 138) ->
 43;
yeccgoto(expr_600, 150) ->
 43;
yeccgoto(expr_600, 155) ->
 43;
yeccgoto(expr_600, 168) ->
 170;
yeccgoto(expr_600, 175) ->
 43;
yeccgoto(expr_600, 180) ->
 43;
yeccgoto(expr_600, 193) ->
 43;
yeccgoto(expr_600, 195) ->
 43;
yeccgoto(expr_600, 197) ->
 43;
yeccgoto(expr_600, 199) ->
 43;
yeccgoto(expr_600, 200) ->
 43;
yeccgoto(expr_600, 203) ->
 43;
yeccgoto(expr_600, 207) ->
 43;
yeccgoto(expr_600, 214) ->
 43;
yeccgoto(expr_600, 217) ->
 43;
yeccgoto(expr_600, 252) ->
 43;
yeccgoto(expr_600, 270) ->
 43;
yeccgoto(expr_700, 14) ->
 44;
yeccgoto(expr_700, 18) ->
 44;
yeccgoto(expr_700, 20) ->
 44;
yeccgoto(expr_700, 25) ->
 44;
yeccgoto(expr_700, 28) ->
 44;
yeccgoto(expr_700, 31) ->
 44;
yeccgoto(expr_700, 33) ->
 44;
yeccgoto(expr_700, 54) ->
 44;
yeccgoto(expr_700, 60) ->
 119;
yeccgoto(expr_700, 63) ->
 44;
yeccgoto(expr_700, 68) ->
 44;
yeccgoto(expr_700, 72) ->
 44;
yeccgoto(expr_700, 77) ->
 44;
yeccgoto(expr_700, 83) ->
 44;
yeccgoto(expr_700, 87) ->
 44;
yeccgoto(expr_700, 90) ->
 44;
yeccgoto(expr_700, 92) ->
 44;
yeccgoto(expr_700, 94) ->
 44;
yeccgoto(expr_700, 99) ->
 44;
yeccgoto(expr_700, 106) ->
 44;
yeccgoto(expr_700, 110) ->
 44;
yeccgoto(expr_700, 115) ->
 44;
yeccgoto(expr_700, 117) ->
 44;
yeccgoto(expr_700, 124) ->
 44;
yeccgoto(expr_700, 138) ->
 44;
yeccgoto(expr_700, 150) ->
 44;
yeccgoto(expr_700, 155) ->
 44;
yeccgoto(expr_700, 168) ->
 44;
yeccgoto(expr_700, 175) ->
 44;
yeccgoto(expr_700, 180) ->
 44;
yeccgoto(expr_700, 193) ->
 44;
yeccgoto(expr_700, 195) ->
 44;
yeccgoto(expr_700, 197) ->
 44;
yeccgoto(expr_700, 199) ->
 44;
yeccgoto(expr_700, 200) ->
 44;
yeccgoto(expr_700, 203) ->
 44;
yeccgoto(expr_700, 207) ->
 44;
yeccgoto(expr_700, 214) ->
 44;
yeccgoto(expr_700, 217) ->
 44;
yeccgoto(expr_700, 252) ->
 44;
yeccgoto(expr_700, 270) ->
 44;
yeccgoto(expr_800, 14) ->
 45;
yeccgoto(expr_800, 18) ->
 45;
yeccgoto(expr_800, 20) ->
 45;
yeccgoto(expr_800, 25) ->
 45;
yeccgoto(expr_800, 28) ->
 45;
yeccgoto(expr_800, 31) ->
 45;
yeccgoto(expr_800, 33) ->
 45;
yeccgoto(expr_800, 54) ->
 45;
yeccgoto(expr_800, 60) ->
 45;
yeccgoto(expr_800, 63) ->
 45;
yeccgoto(expr_800, 68) ->
 45;
yeccgoto(expr_800, 72) ->
 45;
yeccgoto(expr_800, 77) ->
 45;
yeccgoto(expr_800, 83) ->
 45;
yeccgoto(expr_800, 87) ->
 45;
yeccgoto(expr_800, 90) ->
 45;
yeccgoto(expr_800, 92) ->
 45;
yeccgoto(expr_800, 94) ->
 45;
yeccgoto(expr_800, 99) ->
 45;
yeccgoto(expr_800, 106) ->
 45;
yeccgoto(expr_800, 110) ->
 45;
yeccgoto(expr_800, 115) ->
 45;
yeccgoto(expr_800, 117) ->
 45;
yeccgoto(expr_800, 124) ->
 45;
yeccgoto(expr_800, 138) ->
 45;
yeccgoto(expr_800, 150) ->
 45;
yeccgoto(expr_800, 155) ->
 45;
yeccgoto(expr_800, 168) ->
 45;
yeccgoto(expr_800, 175) ->
 45;
yeccgoto(expr_800, 180) ->
 45;
yeccgoto(expr_800, 193) ->
 45;
yeccgoto(expr_800, 195) ->
 45;
yeccgoto(expr_800, 197) ->
 45;
yeccgoto(expr_800, 199) ->
 45;
yeccgoto(expr_800, 200) ->
 45;
yeccgoto(expr_800, 203) ->
 45;
yeccgoto(expr_800, 207) ->
 45;
yeccgoto(expr_800, 214) ->
 45;
yeccgoto(expr_800, 217) ->
 45;
yeccgoto(expr_800, 252) ->
 45;
yeccgoto(expr_800, 270) ->
 45;
yeccgoto(expr_900, 14) ->
 46;
yeccgoto(expr_900, 18) ->
 46;
yeccgoto(expr_900, 20) ->
 46;
yeccgoto(expr_900, 25) ->
 46;
yeccgoto(expr_900, 28) ->
 46;
yeccgoto(expr_900, 31) ->
 46;
yeccgoto(expr_900, 33) ->
 46;
yeccgoto(expr_900, 54) ->
 46;
yeccgoto(expr_900, 60) ->
 46;
yeccgoto(expr_900, 63) ->
 46;
yeccgoto(expr_900, 68) ->
 46;
yeccgoto(expr_900, 72) ->
 46;
yeccgoto(expr_900, 77) ->
 46;
yeccgoto(expr_900, 83) ->
 46;
yeccgoto(expr_900, 87) ->
 46;
yeccgoto(expr_900, 90) ->
 46;
yeccgoto(expr_900, 92) ->
 46;
yeccgoto(expr_900, 94) ->
 46;
yeccgoto(expr_900, 99) ->
 46;
yeccgoto(expr_900, 106) ->
 46;
yeccgoto(expr_900, 110) ->
 46;
yeccgoto(expr_900, 115) ->
 46;
yeccgoto(expr_900, 117) ->
 46;
yeccgoto(expr_900, 124) ->
 46;
yeccgoto(expr_900, 138) ->
 46;
yeccgoto(expr_900, 150) ->
 46;
yeccgoto(expr_900, 155) ->
 46;
yeccgoto(expr_900, 168) ->
 46;
yeccgoto(expr_900, 175) ->
 46;
yeccgoto(expr_900, 180) ->
 46;
yeccgoto(expr_900, 193) ->
 46;
yeccgoto(expr_900, 195) ->
 46;
yeccgoto(expr_900, 197) ->
 46;
yeccgoto(expr_900, 199) ->
 46;
yeccgoto(expr_900, 200) ->
 46;
yeccgoto(expr_900, 203) ->
 46;
yeccgoto(expr_900, 207) ->
 46;
yeccgoto(expr_900, 214) ->
 46;
yeccgoto(expr_900, 217) ->
 46;
yeccgoto(expr_900, 252) ->
 46;
yeccgoto(expr_900, 270) ->
 46;
yeccgoto(expr_max, 14) ->
 47;
yeccgoto(expr_max, 18) ->
 47;
yeccgoto(expr_max, 20) ->
 47;
yeccgoto(expr_max, 24) ->
 226;
yeccgoto(expr_max, 25) ->
 47;
yeccgoto(expr_max, 28) ->
 47;
yeccgoto(expr_max, 31) ->
 47;
yeccgoto(expr_max, 33) ->
 47;
yeccgoto(expr_max, 54) ->
 47;
yeccgoto(expr_max, 60) ->
 47;
yeccgoto(expr_max, 63) ->
 47;
yeccgoto(expr_max, 68) ->
 47;
yeccgoto(expr_max, 72) ->
 47;
yeccgoto(expr_max, 77) ->
 47;
yeccgoto(expr_max, 83) ->
 47;
yeccgoto(expr_max, 87) ->
 47;
yeccgoto(expr_max, 90) ->
 47;
yeccgoto(expr_max, 92) ->
 47;
yeccgoto(expr_max, 94) ->
 47;
yeccgoto(expr_max, 99) ->
 47;
yeccgoto(expr_max, 106) ->
 47;
yeccgoto(expr_max, 110) ->
 47;
yeccgoto(expr_max, 115) ->
 47;
yeccgoto(expr_max, 117) ->
 47;
yeccgoto(expr_max, 124) ->
 47;
yeccgoto(expr_max, 138) ->
 47;
yeccgoto(expr_max, 150) ->
 47;
yeccgoto(expr_max, 155) ->
 47;
yeccgoto(expr_max, 159) ->
 160;
yeccgoto(expr_max, 168) ->
 47;
yeccgoto(expr_max, 175) ->
 47;
yeccgoto(expr_max, 180) ->
 47;
yeccgoto(expr_max, 193) ->
 47;
yeccgoto(expr_max, 195) ->
 47;
yeccgoto(expr_max, 197) ->
 47;
yeccgoto(expr_max, 199) ->
 47;
yeccgoto(expr_max, 200) ->
 47;
yeccgoto(expr_max, 203) ->
 47;
yeccgoto(expr_max, 207) ->
 47;
yeccgoto(expr_max, 214) ->
 47;
yeccgoto(expr_max, 217) ->
 47;
yeccgoto(expr_max, 227) ->
 228;
yeccgoto(expr_max, 229) ->
 241;
yeccgoto(expr_max, 243) ->
 226;
yeccgoto(expr_max, 252) ->
 47;
yeccgoto(expr_max, 270) ->
 47;
yeccgoto(exprs, 14) ->
 256;
yeccgoto(exprs, 18) ->
 48;
yeccgoto(exprs, 28) ->
 210;
yeccgoto(exprs, 54) ->
 48;
yeccgoto(exprs, 68) ->
 76;
yeccgoto(exprs, 72) ->
 73;
yeccgoto(exprs, 83) ->
 85;
yeccgoto(exprs, 124) ->
 48;
yeccgoto(exprs, 138) ->
 48;
yeccgoto(exprs, 203) ->
 204;
yeccgoto(exprs, 270) ->
 272;
yeccgoto(form, 0) ->
 4;
yeccgoto(fun_clause, 50) ->
 129;
yeccgoto(fun_clause, 132) ->
 129;
yeccgoto(fun_clauses, 50) ->
 130;
yeccgoto(fun_clauses, 132) ->
 133;
yeccgoto(fun_expr, 14) ->
 51;
yeccgoto(fun_expr, 18) ->
 51;
yeccgoto(fun_expr, 20) ->
 51;
yeccgoto(fun_expr, 24) ->
 51;
yeccgoto(fun_expr, 25) ->
 51;
yeccgoto(fun_expr, 28) ->
 51;
yeccgoto(fun_expr, 31) ->
 51;
yeccgoto(fun_expr, 33) ->
 51;
yeccgoto(fun_expr, 54) ->
 51;
yeccgoto(fun_expr, 60) ->
 51;
yeccgoto(fun_expr, 63) ->
 51;
yeccgoto(fun_expr, 68) ->
 51;
yeccgoto(fun_expr, 72) ->
 51;
yeccgoto(fun_expr, 77) ->
 51;
yeccgoto(fun_expr, 83) ->
 51;
yeccgoto(fun_expr, 87) ->
 51;
yeccgoto(fun_expr, 90) ->
 51;
yeccgoto(fun_expr, 92) ->
 51;
yeccgoto(fun_expr, 94) ->
 51;
yeccgoto(fun_expr, 99) ->
 51;
yeccgoto(fun_expr, 106) ->
 51;
yeccgoto(fun_expr, 110) ->
 51;
yeccgoto(fun_expr, 115) ->
 51;
yeccgoto(fun_expr, 117) ->
 51;
yeccgoto(fun_expr, 124) ->
 51;
yeccgoto(fun_expr, 138) ->
 51;
yeccgoto(fun_expr, 150) ->
 51;
yeccgoto(fun_expr, 155) ->
 51;
yeccgoto(fun_expr, 159) ->
 51;
yeccgoto(fun_expr, 168) ->
 51;
yeccgoto(fun_expr, 175) ->
 51;
yeccgoto(fun_expr, 180) ->
 51;
yeccgoto(fun_expr, 193) ->
 51;
yeccgoto(fun_expr, 195) ->
 51;
yeccgoto(fun_expr, 197) ->
 51;
yeccgoto(fun_expr, 199) ->
 51;
yeccgoto(fun_expr, 200) ->
 51;
yeccgoto(fun_expr, 203) ->
 51;
yeccgoto(fun_expr, 207) ->
 51;
yeccgoto(fun_expr, 214) ->
 51;
yeccgoto(fun_expr, 217) ->
 51;
yeccgoto(fun_expr, 227) ->
 51;
yeccgoto(fun_expr, 229) ->
 51;
yeccgoto(fun_expr, 243) ->
 51;
yeccgoto(fun_expr, 252) ->
 51;
yeccgoto(fun_expr, 270) ->
 51;
yeccgoto(function, 0) ->
 5;
yeccgoto(function_call, 14) ->
 52;
yeccgoto(function_call, 18) ->
 52;
yeccgoto(function_call, 20) ->
 52;
yeccgoto(function_call, 25) ->
 52;
yeccgoto(function_call, 28) ->
 52;
yeccgoto(function_call, 31) ->
 52;
yeccgoto(function_call, 33) ->
 52;
yeccgoto(function_call, 54) ->
 52;
yeccgoto(function_call, 60) ->
 52;
yeccgoto(function_call, 63) ->
 52;
yeccgoto(function_call, 68) ->
 52;
yeccgoto(function_call, 72) ->
 52;
yeccgoto(function_call, 77) ->
 52;
yeccgoto(function_call, 83) ->
 52;
yeccgoto(function_call, 87) ->
 52;
yeccgoto(function_call, 90) ->
 52;
yeccgoto(function_call, 92) ->
 52;
yeccgoto(function_call, 94) ->
 52;
yeccgoto(function_call, 99) ->
 52;
yeccgoto(function_call, 106) ->
 52;
yeccgoto(function_call, 110) ->
 52;
yeccgoto(function_call, 115) ->
 52;
yeccgoto(function_call, 117) ->
 52;
yeccgoto(function_call, 124) ->
 52;
yeccgoto(function_call, 138) ->
 52;
yeccgoto(function_call, 150) ->
 52;
yeccgoto(function_call, 155) ->
 52;
yeccgoto(function_call, 168) ->
 52;
yeccgoto(function_call, 175) ->
 52;
yeccgoto(function_call, 180) ->
 52;
yeccgoto(function_call, 193) ->
 52;
yeccgoto(function_call, 195) ->
 52;
yeccgoto(function_call, 197) ->
 52;
yeccgoto(function_call, 199) ->
 52;
yeccgoto(function_call, 200) ->
 52;
yeccgoto(function_call, 203) ->
 52;
yeccgoto(function_call, 207) ->
 52;
yeccgoto(function_call, 214) ->
 52;
yeccgoto(function_call, 217) ->
 52;
yeccgoto(function_call, 252) ->
 52;
yeccgoto(function_call, 270) ->
 52;
yeccgoto(function_clause, 0) ->
 6;
yeccgoto(function_clause, 259) ->
 6;
yeccgoto(function_clauses, 0) ->
 7;
yeccgoto(function_clauses, 259) ->
 261;
yeccgoto(guard, 18) ->
 53;
yeccgoto(guard, 54) ->
 120;
yeccgoto(guard, 124) ->
 120;
yeccgoto(guard, 138) ->
 139;
yeccgoto(if_clause, 54) ->
 121;
yeccgoto(if_clause, 124) ->
 121;
yeccgoto(if_clauses, 54) ->
 122;
yeccgoto(if_clauses, 124) ->
 125;
yeccgoto(if_expr, 14) ->
 55;
yeccgoto(if_expr, 18) ->
 55;
yeccgoto(if_expr, 20) ->
 55;
yeccgoto(if_expr, 24) ->
 55;
yeccgoto(if_expr, 25) ->
 55;
yeccgoto(if_expr, 28) ->
 55;
yeccgoto(if_expr, 31) ->
 55;
yeccgoto(if_expr, 33) ->
 55;
yeccgoto(if_expr, 54) ->
 55;
yeccgoto(if_expr, 60) ->
 55;
yeccgoto(if_expr, 63) ->
 55;
yeccgoto(if_expr, 68) ->
 55;
yeccgoto(if_expr, 72) ->
 55;
yeccgoto(if_expr, 77) ->
 55;
yeccgoto(if_expr, 83) ->
 55;
yeccgoto(if_expr, 87) ->
 55;
yeccgoto(if_expr, 90) ->
 55;
yeccgoto(if_expr, 92) ->
 55;
yeccgoto(if_expr, 94) ->
 55;
yeccgoto(if_expr, 99) ->
 55;
yeccgoto(if_expr, 106) ->
 55;
yeccgoto(if_expr, 110) ->
 55;
yeccgoto(if_expr, 115) ->
 55;
yeccgoto(if_expr, 117) ->
 55;
yeccgoto(if_expr, 124) ->
 55;
yeccgoto(if_expr, 138) ->
 55;
yeccgoto(if_expr, 150) ->
 55;
yeccgoto(if_expr, 155) ->
 55;
yeccgoto(if_expr, 159) ->
 55;
yeccgoto(if_expr, 168) ->
 55;
yeccgoto(if_expr, 175) ->
 55;
yeccgoto(if_expr, 180) ->
 55;
yeccgoto(if_expr, 193) ->
 55;
yeccgoto(if_expr, 195) ->
 55;
yeccgoto(if_expr, 197) ->
 55;
yeccgoto(if_expr, 199) ->
 55;
yeccgoto(if_expr, 200) ->
 55;
yeccgoto(if_expr, 203) ->
 55;
yeccgoto(if_expr, 207) ->
 55;
yeccgoto(if_expr, 214) ->
 55;
yeccgoto(if_expr, 217) ->
 55;
yeccgoto(if_expr, 227) ->
 55;
yeccgoto(if_expr, 229) ->
 55;
yeccgoto(if_expr, 243) ->
 55;
yeccgoto(if_expr, 252) ->
 55;
yeccgoto(if_expr, 270) ->
 55;
yeccgoto(lc_expr, 110) ->
 112;
yeccgoto(lc_expr, 115) ->
 112;
yeccgoto(lc_expr, 252) ->
 112;
yeccgoto(lc_exprs, 110) ->
 113;
yeccgoto(lc_exprs, 115) ->
 116;
yeccgoto(lc_exprs, 252) ->
 254;
yeccgoto(list, 14) ->
 57;
yeccgoto(list, 18) ->
 57;
yeccgoto(list, 20) ->
 57;
yeccgoto(list, 24) ->
 57;
yeccgoto(list, 25) ->
 57;
yeccgoto(list, 28) ->
 57;
yeccgoto(list, 31) ->
 57;
yeccgoto(list, 33) ->
 57;
yeccgoto(list, 54) ->
 57;
yeccgoto(list, 60) ->
 57;
yeccgoto(list, 63) ->
 57;
yeccgoto(list, 68) ->
 57;
yeccgoto(list, 72) ->
 57;
yeccgoto(list, 77) ->
 57;
yeccgoto(list, 83) ->
 57;
yeccgoto(list, 87) ->
 57;
yeccgoto(list, 90) ->
 57;
yeccgoto(list, 92) ->
 57;
yeccgoto(list, 94) ->
 57;
yeccgoto(list, 99) ->
 57;
yeccgoto(list, 106) ->
 57;
yeccgoto(list, 110) ->
 57;
yeccgoto(list, 115) ->
 57;
yeccgoto(list, 117) ->
 57;
yeccgoto(list, 124) ->
 57;
yeccgoto(list, 138) ->
 57;
yeccgoto(list, 150) ->
 57;
yeccgoto(list, 155) ->
 57;
yeccgoto(list, 159) ->
 57;
yeccgoto(list, 168) ->
 57;
yeccgoto(list, 175) ->
 57;
yeccgoto(list, 180) ->
 57;
yeccgoto(list, 193) ->
 57;
yeccgoto(list, 195) ->
 57;
yeccgoto(list, 197) ->
 57;
yeccgoto(list, 199) ->
 57;
yeccgoto(list, 200) ->
 57;
yeccgoto(list, 203) ->
 57;
yeccgoto(list, 207) ->
 57;
yeccgoto(list, 214) ->
 57;
yeccgoto(list, 217) ->
 57;
yeccgoto(list, 227) ->
 57;
yeccgoto(list, 229) ->
 57;
yeccgoto(list, 243) ->
 57;
yeccgoto(list, 252) ->
 57;
yeccgoto(list, 270) ->
 57;
yeccgoto(list_comprehension, 14) ->
 58;
yeccgoto(list_comprehension, 18) ->
 58;
yeccgoto(list_comprehension, 20) ->
 58;
yeccgoto(list_comprehension, 24) ->
 58;
yeccgoto(list_comprehension, 25) ->
 58;
yeccgoto(list_comprehension, 28) ->
 58;
yeccgoto(list_comprehension, 31) ->
 58;
yeccgoto(list_comprehension, 33) ->
 58;
yeccgoto(list_comprehension, 54) ->
 58;
yeccgoto(list_comprehension, 60) ->
 58;
yeccgoto(list_comprehension, 61) ->
 107;
yeccgoto(list_comprehension, 63) ->
 58;
yeccgoto(list_comprehension, 68) ->
 58;
yeccgoto(list_comprehension, 72) ->
 58;
yeccgoto(list_comprehension, 77) ->
 58;
yeccgoto(list_comprehension, 83) ->
 58;
yeccgoto(list_comprehension, 87) ->
 58;
yeccgoto(list_comprehension, 90) ->
 58;
yeccgoto(list_comprehension, 92) ->
 58;
yeccgoto(list_comprehension, 94) ->
 58;
yeccgoto(list_comprehension, 99) ->
 58;
yeccgoto(list_comprehension, 106) ->
 58;
yeccgoto(list_comprehension, 110) ->
 58;
yeccgoto(list_comprehension, 115) ->
 58;
yeccgoto(list_comprehension, 117) ->
 58;
yeccgoto(list_comprehension, 124) ->
 58;
yeccgoto(list_comprehension, 138) ->
 58;
yeccgoto(list_comprehension, 150) ->
 58;
yeccgoto(list_comprehension, 155) ->
 58;
yeccgoto(list_comprehension, 159) ->
 58;
yeccgoto(list_comprehension, 168) ->
 58;
yeccgoto(list_comprehension, 175) ->
 58;
yeccgoto(list_comprehension, 180) ->
 58;
yeccgoto(list_comprehension, 193) ->
 58;
yeccgoto(list_comprehension, 195) ->
 58;
yeccgoto(list_comprehension, 197) ->
 58;
yeccgoto(list_comprehension, 199) ->
 58;
yeccgoto(list_comprehension, 200) ->
 58;
yeccgoto(list_comprehension, 203) ->
 58;
yeccgoto(list_comprehension, 207) ->
 58;
yeccgoto(list_comprehension, 214) ->
 58;
yeccgoto(list_comprehension, 217) ->
 58;
yeccgoto(list_comprehension, 227) ->
 58;
yeccgoto(list_comprehension, 229) ->
 58;
yeccgoto(list_comprehension, 243) ->
 58;
yeccgoto(list_comprehension, 252) ->
 58;
yeccgoto(list_comprehension, 270) ->
 58;
yeccgoto(list_op, 41) ->
 180;
yeccgoto(mult_op, 42) ->
 168;
yeccgoto(mult_op, 184) ->
 168;
yeccgoto(opt_bit_size_expr, 225) ->
 230;
yeccgoto(opt_bit_type_list, 230) ->
 232;
yeccgoto(prefix_op, 14) ->
 60;
yeccgoto(prefix_op, 18) ->
 60;
yeccgoto(prefix_op, 20) ->
 60;
yeccgoto(prefix_op, 24) ->
 227;
yeccgoto(prefix_op, 25) ->
 60;
yeccgoto(prefix_op, 28) ->
 60;
yeccgoto(prefix_op, 31) ->
 60;
yeccgoto(prefix_op, 33) ->
 60;
yeccgoto(prefix_op, 54) ->
 60;
yeccgoto(prefix_op, 63) ->
 60;
yeccgoto(prefix_op, 68) ->
 60;
yeccgoto(prefix_op, 72) ->
 60;
yeccgoto(prefix_op, 77) ->
 60;
yeccgoto(prefix_op, 83) ->
 60;
yeccgoto(prefix_op, 87) ->
 60;
yeccgoto(prefix_op, 90) ->
 60;
yeccgoto(prefix_op, 92) ->
 60;
yeccgoto(prefix_op, 94) ->
 60;
yeccgoto(prefix_op, 99) ->
 60;
yeccgoto(prefix_op, 106) ->
 60;
yeccgoto(prefix_op, 110) ->
 60;
yeccgoto(prefix_op, 115) ->
 60;
yeccgoto(prefix_op, 117) ->
 60;
yeccgoto(prefix_op, 124) ->
 60;
yeccgoto(prefix_op, 138) ->
 60;
yeccgoto(prefix_op, 150) ->
 60;
yeccgoto(prefix_op, 155) ->
 60;
yeccgoto(prefix_op, 168) ->
 60;
yeccgoto(prefix_op, 175) ->
 60;
yeccgoto(prefix_op, 180) ->
 60;
yeccgoto(prefix_op, 193) ->
 60;
yeccgoto(prefix_op, 195) ->
 60;
yeccgoto(prefix_op, 197) ->
 60;
yeccgoto(prefix_op, 199) ->
 60;
yeccgoto(prefix_op, 200) ->
 60;
yeccgoto(prefix_op, 203) ->
 60;
yeccgoto(prefix_op, 207) ->
 60;
yeccgoto(prefix_op, 214) ->
 60;
yeccgoto(prefix_op, 217) ->
 60;
yeccgoto(prefix_op, 243) ->
 227;
yeccgoto(prefix_op, 252) ->
 60;
yeccgoto(prefix_op, 270) ->
 60;
yeccgoto(query_expr, 14) ->
 62;
yeccgoto(query_expr, 18) ->
 62;
yeccgoto(query_expr, 20) ->
 62;
yeccgoto(query_expr, 24) ->
 62;
yeccgoto(query_expr, 25) ->
 62;
yeccgoto(query_expr, 28) ->
 62;
yeccgoto(query_expr, 31) ->
 62;
yeccgoto(query_expr, 33) ->
 62;
yeccgoto(query_expr, 54) ->
 62;
yeccgoto(query_expr, 60) ->
 62;
yeccgoto(query_expr, 63) ->
 62;
yeccgoto(query_expr, 68) ->
 62;
yeccgoto(query_expr, 72) ->
 62;
yeccgoto(query_expr, 77) ->
 62;
yeccgoto(query_expr, 83) ->
 62;
yeccgoto(query_expr, 87) ->
 62;
yeccgoto(query_expr, 90) ->
 62;
yeccgoto(query_expr, 92) ->
 62;
yeccgoto(query_expr, 94) ->
 62;
yeccgoto(query_expr, 99) ->
 62;
yeccgoto(query_expr, 106) ->
 62;
yeccgoto(query_expr, 110) ->
 62;
yeccgoto(query_expr, 115) ->
 62;
yeccgoto(query_expr, 117) ->
 62;
yeccgoto(query_expr, 124) ->
 62;
yeccgoto(query_expr, 138) ->
 62;
yeccgoto(query_expr, 150) ->
 62;
yeccgoto(query_expr, 155) ->
 62;
yeccgoto(query_expr, 159) ->
 62;
yeccgoto(query_expr, 168) ->
 62;
yeccgoto(query_expr, 175) ->
 62;
yeccgoto(query_expr, 180) ->
 62;
yeccgoto(query_expr, 193) ->
 62;
yeccgoto(query_expr, 195) ->
 62;
yeccgoto(query_expr, 197) ->
 62;
yeccgoto(query_expr, 199) ->
 62;
yeccgoto(query_expr, 200) ->
 62;
yeccgoto(query_expr, 203) ->
 62;
yeccgoto(query_expr, 207) ->
 62;
yeccgoto(query_expr, 214) ->
 62;
yeccgoto(query_expr, 217) ->
 62;
yeccgoto(query_expr, 227) ->
 62;
yeccgoto(query_expr, 229) ->
 62;
yeccgoto(query_expr, 243) ->
 62;
yeccgoto(query_expr, 252) ->
 62;
yeccgoto(query_expr, 270) ->
 62;
yeccgoto(receive_expr, 14) ->
 64;
yeccgoto(receive_expr, 18) ->
 64;
yeccgoto(receive_expr, 20) ->
 64;
yeccgoto(receive_expr, 24) ->
 64;
yeccgoto(receive_expr, 25) ->
 64;
yeccgoto(receive_expr, 28) ->
 64;
yeccgoto(receive_expr, 31) ->
 64;
yeccgoto(receive_expr, 33) ->
 64;
yeccgoto(receive_expr, 54) ->
 64;
yeccgoto(receive_expr, 60) ->
 64;
yeccgoto(receive_expr, 63) ->
 64;
yeccgoto(receive_expr, 68) ->
 64;
yeccgoto(receive_expr, 72) ->
 64;
yeccgoto(receive_expr, 77) ->
 64;
yeccgoto(receive_expr, 83) ->
 64;
yeccgoto(receive_expr, 87) ->
 64;
yeccgoto(receive_expr, 90) ->
 64;
yeccgoto(receive_expr, 92) ->
 64;
yeccgoto(receive_expr, 94) ->
 64;
yeccgoto(receive_expr, 99) ->
 64;
yeccgoto(receive_expr, 106) ->
 64;
yeccgoto(receive_expr, 110) ->
 64;
yeccgoto(receive_expr, 115) ->
 64;
yeccgoto(receive_expr, 117) ->
 64;
yeccgoto(receive_expr, 124) ->
 64;
yeccgoto(receive_expr, 138) ->
 64;
yeccgoto(receive_expr, 150) ->
 64;
yeccgoto(receive_expr, 155) ->
 64;
yeccgoto(receive_expr, 159) ->
 64;
yeccgoto(receive_expr, 168) ->
 64;
yeccgoto(receive_expr, 175) ->
 64;
yeccgoto(receive_expr, 180) ->
 64;
yeccgoto(receive_expr, 193) ->
 64;
yeccgoto(receive_expr, 195) ->
 64;
yeccgoto(receive_expr, 197) ->
 64;
yeccgoto(receive_expr, 199) ->
 64;
yeccgoto(receive_expr, 200) ->
 64;
yeccgoto(receive_expr, 203) ->
 64;
yeccgoto(receive_expr, 207) ->
 64;
yeccgoto(receive_expr, 214) ->
 64;
yeccgoto(receive_expr, 217) ->
 64;
yeccgoto(receive_expr, 227) ->
 64;
yeccgoto(receive_expr, 229) ->
 64;
yeccgoto(receive_expr, 243) ->
 64;
yeccgoto(receive_expr, 252) ->
 64;
yeccgoto(receive_expr, 270) ->
 64;
yeccgoto(record_expr, 14) ->
 65;
yeccgoto(record_expr, 18) ->
 65;
yeccgoto(record_expr, 20) ->
 65;
yeccgoto(record_expr, 25) ->
 65;
yeccgoto(record_expr, 28) ->
 65;
yeccgoto(record_expr, 31) ->
 65;
yeccgoto(record_expr, 33) ->
 65;
yeccgoto(record_expr, 54) ->
 65;
yeccgoto(record_expr, 60) ->
 65;
yeccgoto(record_expr, 63) ->
 65;
yeccgoto(record_expr, 68) ->
 65;
yeccgoto(record_expr, 72) ->
 65;
yeccgoto(record_expr, 77) ->
 65;
yeccgoto(record_expr, 83) ->
 65;
yeccgoto(record_expr, 87) ->
 65;
yeccgoto(record_expr, 90) ->
 65;
yeccgoto(record_expr, 92) ->
 65;
yeccgoto(record_expr, 94) ->
 65;
yeccgoto(record_expr, 99) ->
 65;
yeccgoto(record_expr, 106) ->
 65;
yeccgoto(record_expr, 110) ->
 65;
yeccgoto(record_expr, 115) ->
 65;
yeccgoto(record_expr, 117) ->
 65;
yeccgoto(record_expr, 124) ->
 65;
yeccgoto(record_expr, 138) ->
 65;
yeccgoto(record_expr, 150) ->
 65;
yeccgoto(record_expr, 155) ->
 65;
yeccgoto(record_expr, 168) ->
 65;
yeccgoto(record_expr, 175) ->
 65;
yeccgoto(record_expr, 180) ->
 65;
yeccgoto(record_expr, 193) ->
 65;
yeccgoto(record_expr, 195) ->
 65;
yeccgoto(record_expr, 197) ->
 65;
yeccgoto(record_expr, 199) ->
 65;
yeccgoto(record_expr, 200) ->
 65;
yeccgoto(record_expr, 203) ->
 65;
yeccgoto(record_expr, 207) ->
 65;
yeccgoto(record_expr, 214) ->
 65;
yeccgoto(record_expr, 217) ->
 65;
yeccgoto(record_expr, 252) ->
 65;
yeccgoto(record_expr, 270) ->
 65;
yeccgoto(record_field, 144) ->
 146;
yeccgoto(record_field, 153) ->
 146;
yeccgoto(record_fields, 144) ->
 147;
yeccgoto(record_fields, 153) ->
 154;
yeccgoto(record_tuple, 141) ->
 143;
yeccgoto(record_tuple, 248) ->
 250;
yeccgoto(rule, 0) ->
 8;
yeccgoto(rule_body, 17) ->
 253;
yeccgoto(rule_body, 268) ->
 253;
yeccgoto(rule_clause, 0) ->
 9;
yeccgoto(rule_clause, 11) ->
 9;
yeccgoto(rule_clauses, 0) ->
 10;
yeccgoto(rule_clauses, 11) ->
 13;
yeccgoto(strings, 14) ->
 67;
yeccgoto(strings, 18) ->
 67;
yeccgoto(strings, 20) ->
 67;
yeccgoto(strings, 24) ->
 67;
yeccgoto(strings, 25) ->
 67;
yeccgoto(strings, 28) ->
 67;
yeccgoto(strings, 31) ->
 67;
yeccgoto(strings, 33) ->
 67;
yeccgoto(strings, 54) ->
 67;
yeccgoto(strings, 60) ->
 67;
yeccgoto(strings, 63) ->
 67;
yeccgoto(strings, 66) ->
 89;
yeccgoto(strings, 68) ->
 67;
yeccgoto(strings, 72) ->
 67;
yeccgoto(strings, 77) ->
 67;
yeccgoto(strings, 83) ->
 67;
yeccgoto(strings, 87) ->
 67;
yeccgoto(strings, 90) ->
 67;
yeccgoto(strings, 92) ->
 67;
yeccgoto(strings, 94) ->
 67;
yeccgoto(strings, 99) ->
 67;
yeccgoto(strings, 106) ->
 67;
yeccgoto(strings, 110) ->
 67;
yeccgoto(strings, 115) ->
 67;
yeccgoto(strings, 117) ->
 67;
yeccgoto(strings, 124) ->
 67;
yeccgoto(strings, 138) ->
 67;
yeccgoto(strings, 150) ->
 67;
yeccgoto(strings, 155) ->
 67;
yeccgoto(strings, 159) ->
 67;
yeccgoto(strings, 168) ->
 67;
yeccgoto(strings, 175) ->
 67;
yeccgoto(strings, 180) ->
 67;
yeccgoto(strings, 193) ->
 67;
yeccgoto(strings, 195) ->
 67;
yeccgoto(strings, 197) ->
 67;
yeccgoto(strings, 199) ->
 67;
yeccgoto(strings, 200) ->
 67;
yeccgoto(strings, 203) ->
 67;
yeccgoto(strings, 207) ->
 67;
yeccgoto(strings, 214) ->
 67;
yeccgoto(strings, 217) ->
 67;
yeccgoto(strings, 227) ->
 67;
yeccgoto(strings, 229) ->
 67;
yeccgoto(strings, 243) ->
 67;
yeccgoto(strings, 252) ->
 67;
yeccgoto(strings, 270) ->
 67;
yeccgoto(tail, 213) ->
 216;
yeccgoto(tail, 220) ->
 221;
yeccgoto(try_expr, 14) ->
 69;
yeccgoto(try_expr, 18) ->
 69;
yeccgoto(try_expr, 20) ->
 69;
yeccgoto(try_expr, 24) ->
 69;
yeccgoto(try_expr, 25) ->
 69;
yeccgoto(try_expr, 28) ->
 69;
yeccgoto(try_expr, 31) ->
 69;
yeccgoto(try_expr, 33) ->
 69;
yeccgoto(try_expr, 54) ->
 69;
yeccgoto(try_expr, 60) ->
 69;
yeccgoto(try_expr, 63) ->
 69;
yeccgoto(try_expr, 68) ->
 69;
yeccgoto(try_expr, 72) ->
 69;
yeccgoto(try_expr, 77) ->
 69;
yeccgoto(try_expr, 83) ->
 69;
yeccgoto(try_expr, 87) ->
 69;
yeccgoto(try_expr, 90) ->
 69;
yeccgoto(try_expr, 92) ->
 69;
yeccgoto(try_expr, 94) ->
 69;
yeccgoto(try_expr, 99) ->
 69;
yeccgoto(try_expr, 106) ->
 69;
yeccgoto(try_expr, 110) ->
 69;
yeccgoto(try_expr, 115) ->
 69;
yeccgoto(try_expr, 117) ->
 69;
yeccgoto(try_expr, 124) ->
 69;
yeccgoto(try_expr, 138) ->
 69;
yeccgoto(try_expr, 150) ->
 69;
yeccgoto(try_expr, 155) ->
 69;
yeccgoto(try_expr, 159) ->
 69;
yeccgoto(try_expr, 168) ->
 69;
yeccgoto(try_expr, 175) ->
 69;
yeccgoto(try_expr, 180) ->
 69;
yeccgoto(try_expr, 193) ->
 69;
yeccgoto(try_expr, 195) ->
 69;
yeccgoto(try_expr, 197) ->
 69;
yeccgoto(try_expr, 199) ->
 69;
yeccgoto(try_expr, 200) ->
 69;
yeccgoto(try_expr, 203) ->
 69;
yeccgoto(try_expr, 207) ->
 69;
yeccgoto(try_expr, 214) ->
 69;
yeccgoto(try_expr, 217) ->
 69;
yeccgoto(try_expr, 227) ->
 69;
yeccgoto(try_expr, 229) ->
 69;
yeccgoto(try_expr, 243) ->
 69;
yeccgoto(try_expr, 252) ->
 69;
yeccgoto(try_expr, 270) ->
 69;
yeccgoto(tuple, 14) ->
 70;
yeccgoto(tuple, 18) ->
 70;
yeccgoto(tuple, 20) ->
 70;
yeccgoto(tuple, 24) ->
 70;
yeccgoto(tuple, 25) ->
 70;
yeccgoto(tuple, 28) ->
 70;
yeccgoto(tuple, 31) ->
 70;
yeccgoto(tuple, 33) ->
 70;
yeccgoto(tuple, 54) ->
 70;
yeccgoto(tuple, 60) ->
 70;
yeccgoto(tuple, 63) ->
 70;
yeccgoto(tuple, 68) ->
 70;
yeccgoto(tuple, 72) ->
 70;
yeccgoto(tuple, 77) ->
 70;
yeccgoto(tuple, 83) ->
 70;
yeccgoto(tuple, 87) ->
 70;
yeccgoto(tuple, 90) ->
 70;
yeccgoto(tuple, 92) ->
 70;
yeccgoto(tuple, 94) ->
 70;
yeccgoto(tuple, 99) ->
 70;
yeccgoto(tuple, 106) ->
 70;
yeccgoto(tuple, 110) ->
 70;
yeccgoto(tuple, 115) ->
 70;
yeccgoto(tuple, 117) ->
 70;
yeccgoto(tuple, 124) ->
 70;
yeccgoto(tuple, 138) ->
 70;
yeccgoto(tuple, 150) ->
 70;
yeccgoto(tuple, 155) ->
 70;
yeccgoto(tuple, 159) ->
 70;
yeccgoto(tuple, 168) ->
 70;
yeccgoto(tuple, 175) ->
 70;
yeccgoto(tuple, 180) ->
 70;
yeccgoto(tuple, 193) ->
 70;
yeccgoto(tuple, 195) ->
 70;
yeccgoto(tuple, 197) ->
 70;
yeccgoto(tuple, 199) ->
 70;
yeccgoto(tuple, 200) ->
 70;
yeccgoto(tuple, 203) ->
 70;
yeccgoto(tuple, 207) ->
 70;
yeccgoto(tuple, 214) ->
 70;
yeccgoto(tuple, 217) ->
 70;
yeccgoto(tuple, 227) ->
 70;
yeccgoto(tuple, 229) ->
 70;
yeccgoto(tuple, 243) ->
 70;
yeccgoto(tuple, 252) ->
 70;
yeccgoto(tuple, 270) ->
 70;
yeccgoto(__Symbol, __State) ->
 exit({__Symbol, __State, missing_in_goto_table}).


